@@ -56,10 +56,54 @@
 
 (defvar ement-room-buffer-name-prefix)
 (defvar ement-room-buffer-name-suffix)
+(defvar ement-room-leave-kill-buffer)
 (defvar ement-room-prism)
 (defvar ement-room-prism-color-adjustment)
 (defvar ement-room-prism-minimum-contrast)
 
+;;;; Compatibility
+
+;; These workarounds should be removed when they aren't needed.
+
+;;;;; Emacs 28 color features.
+
+;; Copied from Emacs 28.  See <https://github.com/alphapapa/ement.el/issues/99>.
+
+;; FIXME: Remove this workaround when possible.
+
+(eval-and-compile
+  (unless (boundp 'color-luminance-dark-limit)
+    (defconst ement--color-luminance-dark-limit 0.325
+      "The relative luminance below which a color is considered 'dark'.
+A 'dark' color in this sense provides better contrast with white
+than with black; see `color-dark-p'.
+This value was determined experimentally.")))
+
+(defalias 'ement--color-dark-p
+  (if (fboundp 'color-dark-p)
+      'color-dark-p
+    (lambda (rgb)
+      "Whether RGB is more readable against white than black.
+RGB is a 3-element list (R G B), each component in the range [0,1].
+This predicate can be used both for determining a suitable (black or white)
+contrast colour with RGB as background and as foreground."
+      (unless (<= 0 (apply #'min rgb) (apply #'max rgb) 1)
+        (error "RGB components %S not in [0,1]" rgb))
+      ;; Compute the relative luminance after gamma-correcting (assuming sRGB),
+      ;; and compare to a cut-off value determined experimentally.
+      ;; See https://en.wikipedia.org/wiki/Relative_luminance for details.
+      (let* ((sr (nth 0 rgb))
+             (sg (nth 1 rgb))
+             (sb (nth 2 rgb))
+             ;; Gamma-correct the RGB components to linear values.
+             ;; Use the power 2.2 as an approximation to sRGB gamma;
+             ;; it should be good enough for the purpose of this function.
+             (r (expt sr 2.2))
+             (g (expt sg 2.2))
+             (b (expt sb 2.2))
+             (y (+ (* r 0.2126) (* g 0.7152) (* b 0.0722))))
+        (< y ement--color-luminance-dark-limit)))))
+
 ;;;; Functions
 
 ;;;;; Commands
@@ -100,21 +144,81 @@ the request."
       (ement-api session endpoint :method 'post :data (json-encode data)
         :then then))))
 
-(defun ement-forget-room (room session)
-  "Forget ROOM on SESSION."
-  (interactive (ement-complete-room))
-  (pcase-let* (((cl-struct ement-room id display-name) room)
-               (endpoint (format "rooms/%s/forget" (url-hexify-string id))))
-    (when (yes-or-no-p (format "Forget room \"%s\" (%s)? " display-name id))
+(defun ement-room-leave (room session &optional force-p)
+  "Leave ROOM on SESSION.
+If FORCE-P, leave without prompting.  ROOM may be an `ement-room'
+struct, or a room ID or alias string."
+  ;; TODO: Rename `room' argument to `room-or-id'.
+  (interactive (ement-complete-room :session (ement-complete-session)))
+  (cl-assert room) (cl-assert session)
+  (cl-etypecase room
+    (ement-room)
+    (string (setf room (ement-afirst (or (equal room (ement-room-canonical-alias it))
+                                         (equal room (ement-room-id it)))
+                         (ement-session-rooms session)))))
+  (when (or force-p (yes-or-no-p (format "Leave room %s? " (ement--format-room room))))
+    (pcase-let* (((cl-struct ement-room id) room)
+                 (endpoint (format "rooms/%s/leave" (url-hexify-string id))))
       (ement-api session endpoint :method 'post :data ""
         :then (lambda (_data)
-                ;; NOTE: The spec does not seem to indicate that the action of forgetting
-                ;; a room is synced to other clients, so it seems that we need to remove
-                ;; the room from the session here.
-                (setf (ement-session-rooms session)
-                      (cl-remove room (ement-session-rooms session)))
-                ;; TODO: Indicate forgotten in footer in room buffer.
-                (message "Room \"%s\" (%s) forgotten." display-name id))))))
+                (when ement-room-leave-kill-buffer
+                  ;; NOTE: This generates a symbol and sets its function value to a lambda
+                  ;; which removes the symbol from the hook, removing itself from the hook.
+                  ;; TODO: When requiring Emacs 27, use `letrec'.
+                  (let* ((leave-fn-symbol (gensym (format "ement-leave-%s" room)))
+                         (leave-fn (lambda (_session)
+                                     (remove-hook 'ement-sync-callback-hook leave-fn-symbol)
+                                     ;; FIXME: Probably need to unintern the symbol.
+                                     (when-let ((buffer (map-elt (ement-room-local room) 'buffer)))
+                                       (when (buffer-live-p buffer)
+                                         (kill-buffer buffer))))))
+                    (setf (symbol-function leave-fn-symbol) leave-fn)
+                    (add-hook 'ement-sync-callback-hook leave-fn-symbol)))
+                (message "Left room: %s" (ement--format-room room)))
+        :else (lambda (plz-error)
+                (pcase-let* (((cl-struct plz-error response) plz-error)
+                             ((cl-struct plz-response status body) response)
+                             ((map error) (json-read-from-string body)))
+                  (pcase status
+                    (429 (error "Unable to leave room %s: %s" room error))
+                    (_ (error "Unable to leave room %s: %s %S" room status plz-error)))))))))
+(defalias 'ement-leave-room #'ement-room-leave)
+
+(defun ement-forget-room (room session &optional force-p)
+  "Forget ROOM on SESSION.
+If FORCE-P (interactively, with prefix), prompt to leave the room
+when necessary, and forget the room without prompting."
+  (interactive (pcase-let ((`(,room ,session) (ement-complete-room)))
+                 (list room session current-prefix-arg)))
+  (pcase-let* (((cl-struct ement-room id display-name status) room)
+               (endpoint (format "rooms/%s/forget" (url-hexify-string id))))
+    (pcase status
+      ('join (if (and force-p
+                      (yes-or-no-p (format "Leave and forget room %s? (WARNING: You will not be able to rejoin the room to access its content.) "
+                                           (ement--format-room room))))
+                 (progn
+                   ;; TODO: Use `letrec'.
+                   (let* ((forget-fn-symbol (gensym (format "ement-forget-%s" room)))
+                          (forget-fn (lambda (_session)
+                                       (when (equal 'leave (ement-room-status room))
+                                         (remove-hook 'ement-sync-callback-hook forget-fn-symbol)
+                                         ;; FIXME: Probably need to unintern the symbol.
+                                         (ement-forget-room room session 'force)))))
+                     (setf (symbol-function forget-fn-symbol) forget-fn)
+                     (add-hook 'ement-sync-callback-hook forget-fn-symbol))
+                   (ement-leave-room room session 'force))
+               (user-error "Room %s is joined (must be left before forgetting)"
+                           (ement--format-room room))))
+      ('leave (when (or force-p (yes-or-no-p (format "Forget room \"%s\" (%s)? " display-name id)))
+                (ement-api session endpoint :method 'post :data ""
+                  :then (lambda (_data)
+                          ;; NOTE: The spec does not seem to indicate that the action of forgetting
+                          ;; a room is synced to other clients, so it seems that we need to remove
+                          ;; the room from the session here.
+                          (setf (ement-session-rooms session)
+                                (cl-remove room (ement-session-rooms session)))
+                          ;; TODO: Indicate forgotten in footer in room buffer.
+                          (message "Room \"%s\" (%s) forgotten." display-name id))))))))
 
 (defun ement-ignore-user (user-id session &optional unignore-p)
   "Ignore USER-ID on SESSION.
@@ -230,7 +334,7 @@ members, show in a new buffer; otherwise show in echo area."
                       (heading "Avatar: ") (or avatar "") "\n\n"
                       (heading "ID: ") "<" (id room-id) ">" "\n"
                       (heading "Alias: ") "<" (id canonical-alias) ">" "\n\n"
-                      (heading "Topic: ") (propertize topic 'face 'font-lock-comment-face) "\n\n"
+                      (heading "Topic: ") (propertize (or topic "[none]") 'face 'font-lock-comment-face) "\n\n"
                       (heading "Retrieved events: ") (number-to-string (length timeline)) "\n"
                       (heading "  spanning: ")
                       (format-time-string "%Y-%m-%d %H:%M:%S"
@@ -319,6 +423,11 @@ If DELETE (interactively, with prefix), delete it."
 
 ;; These functions could reasonably be called by code in other packages.
 
+(defun ement-message (format-string &rest args)
+  "Call `message' on FORMAT-STRING prefixed with \"Ement: \"."
+  ;; TODO: Use this function everywhere we use `message'.
+  (apply #'message (concat "Ement: " format-string) args))
+
 (cl-defun ement-upload (session &key data filename then else
                                 (content-type "application/octet-stream"))
   "Upload DATA with FILENAME to content repository on SESSION.
@@ -331,12 +440,12 @@ THEN and ELSE are passed to `ement-api', which see."
     :content-type content-type :data data :data-type 'binary
     :then then :else else))
 
-(defun ement-complete-session ()
+(cl-defun ement-complete-session (&key (prompt "Session: "))
   "Return an Ement session selected with completion."
   (cl-etypecase (length ement-sessions)
     ((integer 1 1) (cdar ement-sessions))
     ((integer 2 *) (let* ((ids (mapcar #'car ement-sessions))
-                          (selected-id (completing-read "Session: " ids nil t)))
+                          (selected-id (completing-read prompt ids nil t)))
                      (alist-get selected-id ement-sessions nil nil #'equal)))
     (otherwise (user-error "No active sessions.  Call `ement-connect' to log in"))))
 
@@ -490,7 +599,7 @@ avatars, etc."
                                                  "white"))
                                               (_
                                                ;; The `contrast-with` color is usable: test it.
-                                               (if (color-dark-p (color-name-to-rgb contrast-with))
+                                               (if (ement--color-dark-p (color-name-to-rgb contrast-with))
                                                    "white" "black")))))))
       (apply #'color-rgb-to-hex (append color-rgb (list 2))))))
 
@@ -808,14 +917,14 @@ m.replace metadata)."
   "Return ROOM formatted with name, alias, ID, and optionally TOPIC.
 Suitable for use in completion, etc."
   (if topic
-      (format "%s <%s> (<%s>): \"%s\""
+      (format "\"%s\" <%s> (<%s>): \"%s\""
               (or (ement-room-display-name room)
                   (setf (ement-room-display-name room)
                         (ement--room-display-name room)))
               (ement-room-canonical-alias room)
               (ement-room-id room)
               (ement-room-topic room))
-    (format "%s <%s> (<%s>)"
+    (format "\"%s\" <%s> (<%s>)"
             (or (ement-room-display-name room)
                 (setf (ement-room-display-name room)
                       (ement--room-display-name room)))
@@ -1062,7 +1171,7 @@ is not at the latest known message event."
                      ;; the user doesn't miss any messages, but it's unclear whether this
                      ;; is really correct or best.)
                      t)
-                    ((not (equal our-id (ement-user-id (ement-event-sender (car timeline)))))
+                    ((equal our-id (ement-user-id (ement-event-sender (car timeline))))
                      ;; We sent the last event: the room is read.
                      nil)
                     ((and first-counting-event
@@ -1148,12 +1257,14 @@ such when they were created."
                (members (delete-dups (mapcar #'ement-event-sender timeline)))
                (other-users (cl-remove local-user-id members
                                        :key #'ement-user-id :test #'equal))
-               ((cl-struct ement-user (id other-user-id)) (car other-users)))
+               ((cl-struct ement-user (id other-user-id)) (car other-users))
+               ;; The alist keys are MXIDs as symbols.
+               (other-user-id (intern other-user-id))
+               (existing-direct-rooms-for-user (map-elt direct-rooms-account-data-event-content other-user-id)))
     (cl-assert (= 1 (length other-users)))
-    (cl-assert (not (map-elt direct-rooms-account-data-event-content other-user-id)) nil
-               "Ement: User <%s> already has a direct room on session <%s>"
-               other-user-id local-user-id)
-    (setf (map-elt direct-rooms-account-data-event-content other-user-id) (vector room-id))
+    (setf (map-elt direct-rooms-account-data-event-content other-user-id)
+          (cl-coerce (append existing-direct-rooms-for-user (list room-id))
+                     'vector))
     (ement-put-account-data session "m.direct" direct-rooms-account-data-event-content
       :then (lambda (_data)
               (message "Ement: Room <%s> marked as direct for <%s>." room-id other-user-id)))
